(in-package :blog)

(defconstant +calendar-day-names+ '("Monday" "Tuesday" "Wednesday" "Thursday" "Friday" "Saturday" "Sunday"))
(defconstant +calendar-month-names+ '("Jan" "Feb" "Mar" "Apr" "May" "Jun" "Jul" "Aug" "Sep" "Oct" "Nov" "Dec"))

(defun weekday-name (week)
  (subseq (nth week +calendar-day-names+) 0 3))
(defun month-name (month) (nth (1- month) +calendar-month-names+))
(defun get-this-year () (sixth (multiple-value-list (get-decoded-time))))
(defun get-this-month () (fifth (multiple-value-list (get-decoded-time))))

(defun calendar (month year)
  (let ((dtime (encode-universal-time 0 0 12 1 month year)))
    (flet ((next-day ()
	     (multiple-value-bind (se mi ho da mo ye week)
		 (decode-universal-time dtime)
	       (declare (ignorable se mi ho ye))
	       (incf dtime (* 60 60 24))
	       (if (= mo month) (list da week) nil))))
      (loop for d = (next-day) while d collect d))))

(defun format-html-calendar (stream month year fn)
  (let* ((c (calendar month year))
	 (p (loop for i from 0 below (second (car c)) collect (list nil i)))
	 (c (nconc p c)))
    (format stream "<table class=\"calendar\" summary=\"~A ~A\">~%<tr><td colspan=\"7\"><div align=\"center\">~A ~A</div></td></tr>~%" 
	    (month-name month) year
	    (month-name month) year)
    (write-line "<tr>" stream)
    (dotimes (n (length +calendar-day-names+))
      (format stream "<td class=\"~A\">~A</td>" (weekday-name n) (weekday-name n)))
    (write-line "</tr>" stream)
    (loop for (day week) in c and i from 0 do
	  (when (= (mod week 7) 0) (format stream "<tr>"))
	  (format stream "<td class=\"~A\">~A</td>"
		  (weekday-name (mod week 7))
		  (if day (funcall fn year month day) "&nbsp;"))
	  (when (= (mod week 7) 6) (format stream "</tr>~%"))
	  finally
	  (loop repeat (- 6 (mod week 7)) do (write-string "<td>&nbsp;</td>" stream) finally (write-line "</tr>" stream)))
    (format stream "</table>~%")))

(defplugin "/(calendar [month] [year] [:center|:right]) - calendar plugin."
  (lambda (blog req page arg stream)
    (declare (ignorable blog req page arg))
    (cond ((find :center arg) (write-line "<div align=\"center\">" stream))
	  ((find :right arg) (write-line "<div align=\"right\">" stream)))
    (format-html-calendar
     stream
     (or (and (not (keywordp (first arg))) (first arg)) (get-this-month))
     (or (and (not (keywordp (second arg))) (second arg))(second arg) (get-this-year))
     (lambda (year month day)
       (let ((name (format nil "~4,'0D~2,'0D~2,'0D" year month day)))
	 (if (find-page blog :page name)
	     (format nil "<a href=\"~A.html\">~A~2D</a>" name (if (< day 10) "&nbsp" "") day)
	     (format nil "~2D" day)))))
    (cond ((find :center arg) (write-line "</div>" stream))
	  ((find :right arg) (write-line "</div>" stream)))))

