(in-package :blog)

(defplugin "/(search [:all|:pages|:days] [\"keyword\"]+) - search keyword."
  (lambda (blog req page arg stream)
    (declare (ignorable blog req page))
    (cond ((and (null arg) (eq (method-of req) :static)))
	  ((null arg)
	   (write-line "<form class=\"tool\" method=\"post\" action=\"index.html\"><p>" stream)
	   (write-line "<input type=\"text\" size=\"15\" maxlength=\"30\" name=\"search\" value=\"\" /> from" stream)
	   (write-line "<input type=\"radio\" name=\"from\" value=\"all\" checked=\"checked\" />All" stream)
	   (write-line "<input type=\"radio\" name=\"from\" value=\"pages\" />Pages" stream)
	   (write-line "<input type=\"radio\" name=\"from\" value=\"days\" />Days" stream)
	   (write-line "<input type=\"submit\" value=\"search\" /></p></form><br />" stream))
	  ((keywordp (car arg))
	   (let* ((str (with-output-to-string (out)
			 (do ((lst (cdr arg) (cdr lst)))
			     ((null lst))
			   (write-string (url-encode-string (car lst)) out)
			   (when (cdr lst) (write-char #\+ out)))))
		  (url (format nil "search_~A+~A.html"
			       (case (car arg) (:all "all") (:pages "pages") (:days "days"))
			       str)))
	     (format stream "<a href=\"~A\">~A</a>" url str)
	     (when (eq (method-of req) :static)
	       (create-link-page blog req url))))
	  (t
	   (let* ((str (with-output-to-string (out)
			 (do ((lst arg (cdr lst)))
			     ((null lst))
			   (write-string (url-encode-string (car lst)) out)
			   (when (cdr lst) (write-char #\+ out)))))
		  (url (format nil "search_all+~A.html" str)))
	     (format stream "<a href=\"~A\">~A</a>" url str)
	     (when (eq (method-of req) :static)
	       (create-link-page blog req url)))))))

(defun create-link-page (blog req url)
  (let* ((len (length url))
	 (name (and (> len 12)
		    (string=? "search_" url) (string= ".html" (subseq url (- len 5)))
		    (subseq url 0 (- len 5))))
	 (page (and name (find-page blog :page name)))
	 (include (included-of blog)))
    (unless page
      (setf page (make-instance '<wiki-page>
				:root blog
				:path (make-pathname :directory (pathname-directory (src-of blog))
						     :name      name
						     :type      "html")))
      (with-open-file (out (make-pathname :directory (pathname-directory (dest-of blog))
					  :name      (name-of page)
					  :type      "html")
			   :direction :output
			   :if-exists :supersede)
	(render-html blog req out url :search)))
    (setf (included-of blog) include)))
