;;; CLISP (WIN32) hack
;;#+WIN32
#+clisp
(eval-when (:execute :load-toplevel :compile-toplevel)
  (let ((encoding (make-encoding :charset "ISO-8859-1" :line-terminator :unix)))
    (system::set-default-file-encoding encoding)
    (system::set-foreign-encoding encoding)
    (system::set-misc-encoding encoding)))

#+clisp
(defpackage :cl-webapp (:use :cl :ext :socket) (:nicknames :webapp :blog) (:export :serve))
#-clisp
(defpackage :cl-webapp (:use :cl :port) (:nicknames :webapp :blog) (:export :serve))
(in-package :cl-webapp)

(defun url-decode-string (s)
  (with-output-to-string (out)
    (do ((i 0 (1+ i)))
        ((>= i (length s)) out)
      (let ((c (char s i)))
        (case c
          (#\% (setf c (code-char (parse-integer s
						 :start (+ i 1) :end (+ i 3)
						 :radix 16 :junk-allowed t)))
	       (unless (null c) (write-char c out))
	       (incf i 2))
          (#\+ (write-char #\Space out))
          (otherwise (write-char c out)))))))

(defun url-encode-string (str)
  (with-output-to-string (out)
    (loop for c across str do
	  (cond	;; ((eq c #\linefeed) (write-string "%0D%0A" out))
	    ((or (alphanumericp c)
		 (find c '(#\- #\_ #\: #\/ #\.)))
	     (write-char c out))
	    ((eq c #\Space)
	     (write-char #\+ out))
	    (t
	     (format out "%~2,'0x" (char-code c)))))))

(defun url-encode-string-only-display-char (str)
  (with-output-to-string (out)
    (write-string "ID" out)
    (loop for c across str do
	  (cond ((find c '(#\- #\_ #\: #\/ #\. #\linefeed #\return #\space))) ;; skip
		(t
		 ;;(format out "%~2,'0x" (if (> (char-code c) 128) (- (char-code c) 128) (char-code c))))))))
		 (format out "~2,'0x" (if (> (char-code c) 128) (- (char-code c) 128) (char-code c))))))))

(defclass <request> ()
  ((method :accessor method-of :initarg :method :initform nil)
   (header :accessor header-of :initform nil)
   (url    :accessor url-of    :initform nil)
   (param  :accessor param-of   :initform nil)))

(defclass <application> ()
  ((base-url  :accessor base-url-of :initarg :base-url)
   (name :accessor name-of :initarg :name)))

(defclass <server> ()
  ((server-socket :accessor server-socket-of :initform nil)
   (applications  :accessor applications-of  :initform (make-hash-table :test #'equal))
   (default       :accessor default-of       :initform (make-instance '<application>))
   (port          :accessor port-of          :initform nil)
   (request       :accessor request-of       :initform nil)))

(defgeneric param-check (self type param value))
(defgeneric localhost? (self))
(defgeneric parse-http-header (self str))
(defgeneric parse-parameter (self str))
(defgeneric parse-url (self str))
(defgeneric start-server (self port))
(defgeneric request (self s))
(defgeneric parse-request (self s))
(defgeneric dispatch (self s))
(defgeneric debug-show (req s))
(defgeneric page-view (application request s))

(defun param-check-of-lst (lst key value)
  (let* ((lst (assoc key lst :test #'string-equal))
	 (var (second lst)))
    (and var
	 (stringp var)
	 (if (functionp value) (funcall value var) (string-equal var value)))))

(defmethod param-check ((self <request>) (type (eql :header)) key value)
  (param-check-of-lst (header-of self) key value))

(defmethod param-check ((self <request>) (type (eql :param)) key value)
  (param-check-of-lst (param-of self) key value))

(defmethod localhost? ((self <request>))
  (param-check self :header "Host"
	       (lambda (e) (let ((host (subseq e 0 (position #\: e))))
			     (or (string-equal "127.0.0.1" host)
				 (string-equal "localhost" host))))))

(defmethod parse-http-header ((self <request>) str)
  (let* ((p (position #\: str))
	 (end (position-if (complement #'(lambda (c) (or (eql c #\return) (eql c #\linefeed))))
			   str
			   :from-end t))
	 (v (and p
		 (list (url-decode-string (subseq str 0 p))
		       (url-decode-string (subseq str (+ p 2) (and end (1+ end))))))))
    (push v (header-of self))))

(defmethod parse-parameter ((self <request>) str)
  (let ((p1 (position #\= str))
	(p2 (position #\& str)))
    (cond (p1
	   (push (list (url-decode-string (subseq str 0 p1))
		       (url-decode-string (subseq str (1+ p1) p2)))
		 (param-of self))
	   (when p2
	     (parse-parameter self (subseq str (1+ p2)))))
          ((string= str "") nil )
          (t (url-decode-string str)))))

(defmethod parse-url ((self <request>) str)
  (let ((p1 (position #\Space str))
	(p2 (position #\Space str :from-end t)))
    ;;(write-line "[request]") ; for debug
    (write-line str)			; for debug
    (when (and p1 p2 (> p2 p1))
      (let ((method (subseq str 0 p1))
	    (url    (subseq str (+ p1 1) p2))
	    (args   nil))
	(when (and ;;(string= method "GET")
	       (setf p1 (position #\? url)))
	  (setf args (subseq url (1+ p1))
		url  (subseq url 0 p1))
	  (parse-parameter self args))
	(setf (method-of self) method
	      (url-of self)    url)))))

(defmethod start-server ((self <server>) port)
  (unless (server-socket-of self)
    (unwind-protect
	 (progn
	   (setf (server-socket-of self)
		 #+clisp
		 (socket-server port)
		 #-clisp
		 (port:open-socket-server port)
		 (port-of self)          port)
	   (format t "server port: ~A~%" port)
	   (loop
	    (request self
		     #+clisp
		     (socket-accept (server-socket-of self))
		     #-clisp
		     (port:socket-accept (server-socket-of self))
		     )))
      #+clisp(socket-server-close (server-socket-of self))
      #-clisp(port:socket-server-close (server-socket-of self))
      (setf (server-socket-of self) nil))))

(defmethod request ((self <server>) s)
  (unwind-protect
       (progn
	 (parse-request self s)
	 (dispatch self s))
    (finish-output s)
    #+clisp(socket-stream-shutdown s :output)
    ;;(close s)
    ))

(defmethod parse-request ((self <server>) s)
  (let ((request (make-instance '<request>)))
    (setf (request-of self) request)
    ;; parse url & arguments
    (parse-url request (read-line s))
    ;; parse http header
    (loop for line = (read-line s nil :eof)
	  until (<= (length line) 2)
	  do
	  ;; (write-line line) ; for debug
	  (parse-http-header request line))
    ;; parse POSTed data
    (let ((h (assoc "Content-Length" (header-of request) :test #'string-equal)))
      (when h ;;(and h (listen s))
	(let* ((length (parse-integer (second h)))
	       (str (make-string length))
	       (c (peek-char nil s)))
	  (if (or (eql c #\return) (eql c #\linefeed))
	      (read-char s))
	  (read-sequence str s)
	  (parse-parameter request str))))))

;; customize for blog
(defmethod dispatch ((self <server>) s)
  (let* ((request (request-of self)))
    (handler-case
	(page-view (default-of self) request s)
      (error (e) (format t "[error] ~A~%" e)))))

(defmethod debug-show ((req <request>) s)
  (format s "<pre>~%header: ~S~%method: ~S~%url: ~S~%param: ~S~%</pre>~%"
	  (header-of req) (method-of req) (url-of req) (param-of req)))

(defmethod page-view ((application <application>) (request <request>) s)
  (write-line "HTTP/1.0 200 OK" s)
  (write-line "Content-Type: text/html; charset=EUC-JP" s)
  (write-line "" s)
  (write-line "<html><body>" s)
  (write-line "<h1>Webapp for Common Lisp</h1>" s)
  (write-line "</body></html>" s))

(defvar *server* (make-instance '<server>))

(defun serve (&optional (port 8080))
  (start-server *server* port))

(defmacro defwebapp (class name-key name base-url-key base-url &rest option)
  (assert (eq name-key :name))
  (assert (eq base-url-key :base-url))
  `(let ((obj (make-instance ',class :name ,name :base-url ,base-url ,@option)))
    (setf (webapp::default-of webapp::*server*) obj)
    (setf (gethash ,base-url (webapp::applications-of webapp::*server*)) obj)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Utility Functions
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun create-directories (name)
  ;;(format t "[debug] >>> ~A <<<~%" name)
  (loop with path = nil
	for d in (pathname-directory name) do
	;;(format t "[debug] >>> ~A~%" d)
	(if (null path)
	    (setq path (make-pathname :directory (list d)))
	    (setq path (make-pathname :directory (append (pathname-directory path) (list d)))))
	;;(format t "[debug] ~A~%" path)
	#+clisp
	(handler-case (ext:make-dir path) (error (e) (declare (ignore e)) nil))
	#+cmu
	(unix:unix-mkdir (directory-namestring path) #o775)))

(defun split-string (str c &key (start 0))
  (do* ((size (length str))
	(lst nil (cons (subseq str beg end) lst))
	(beg start (1+ end))
	(end (position c str :start beg)
	     (position c str :start (min beg size))))
       ((null end) (nreverse (cons (subseq str beg end) lst)))))

(defun string=? (pat str &aux (len (length pat)))
  (and (>= (length str) len)
       (string= pat (subseq str 0 len))))

(defun read-file-contents (file)
  (with-open-file (s file :direction :input
		     #+clisp :element-type #+clisp 'unsigned-byte)
    (let* ((size (file-length s))
	   #+clisp
	   (buf  (make-array size :element-type 'unsigned-byte))
	   #-clisp
	   (buf  (make-string size)))
      (read-sequence buf s)
      #+clisp
      (ext:convert-string-from-bytes buf 'charset:iso-8859-1)
      #-clisp
      buf)))

(defun formatted-date-of (str)
  (if (not (= (length str) 8))
      str
      (let ((year (subseq str 0 4))
	    (mon  (subseq str 4 6))
	    (date (subseq str 6 8)))
	(format nil "~A ~D, ~D"
		(case (parse-integer mon)
		  (1 "January") (2 "February") (3 "March") (4 "April")
		  (5 "May") (6 "June") (7 "July") (8 "August")
		  (9 "September") (10 "October") (11 "November") (12 "December"))
		date year))))

(defun formatted-time-of (time)
  (multiple-value-bind (second minute hour date month year day-of-week dst-p tz)
      (decode-universal-time time)
    (declare (ignorable day-of-week dst-p))
    (format nil " ~2,'0d:~2,'0d:~2,'0d ~d/~2,'0d/~2,'0d (GMT~@d)"
	    hour
	    minute
	    second
	    year
	    month
	    date
	    (- tz))))

(defun scan-stream (chars stream output dispatch)
  (labels ((find-token (chars stream dispatch)
	     (let* ((c1 (read-char stream)))
	       (cons c1
		     (dolist (possible chars)
		       (when (eql c1 (car possible))
			 (return
			   (if (listp (cadr possible))
			       (find-token
				(cadr possible) stream dispatch)
			       (let ((c2 (peek-char nil stream nil)))
				 (when (eql c2 #\( )
				   (funcall dispatch (cadr possible) 
					    (read-matched-parens stream))
				   (throw 'matched nil)))))))))))
    (handler-case
	(loop
	 (catch 'matched
	   (dolist (i (find-token chars stream dispatch))
	     (funcall output i))))
      (end-of-file (c) (declare (ignore c)) nil))))

(defun read-matched-parens (stream)
  (let ((eof (gensym))
        (nesting 0))
    (with-output-to-string (out)
      (loop
       (let ((c (read-char stream nil eof)))
         (if (eql c eof) (return out))
         (if (eql c #\\)
             (progn
               (write-char c out)
               (setf c (read-char stream nil eof)))
	     (progn
	       (if (eql c #\() (incf nesting))
	       (if (eql c #\)) (decf nesting))))
         (write-char c out)
         (if (eql nesting 0) (return out))
         )))))

;;; wiki format support

(defmacro acond (&rest clauses)
  "Anaphoric cond: use `it' in consequent to refer to result of test." ; LMH
  (if (null clauses)
      nil
      (let ((cl1 (car clauses))
            (sym (gensym)))
        `(let ((,sym ,(car cl1)))
	  (if ,sym
	      (let ((it ,sym))
		(declare (ignorable it)) ; LMH
		,@(cdr cl1))
	      (acond ,@(cdr clauses)))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Weblog Tool (System)
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defclass <file> ()
  ((root :accessor root-of :initarg :root)
   (path :accessor path-of :initarg :path :initform nil)
   (contents :accessor contents-of :initform "Edit This Page.")
   (mtime :accessor mtime-of :initarg :mtime :initform nil)))

(defclass <page> (<file>) ((header :accessor header-of :initform "")
			   (body   :accessor body-of   :initform "")
			   (footer :accessor footer-of :initform "")
			   (option :accessor option-of :initform nil)))

(defclass <logfile> (<page>) ())

(defclass <wiki-page> (<page>) ())

(defclass <day-page> (<wiki-page>) ())

(defclass <plugin> (<file>)
  ((document :accessor document-of :initform nil)
   (function :accessor function-of :initform nil)))

(defclass <weblog> (<application>)
  ( ;; configure
   (src    :accessor src-of   :initform "data/")
   (dest   :accessor dest-of   :initform (merge-pathnames "public_html/xrea/" (user-homedir-pathname)))
   (title  :accessor title-of :initform "Lispy Days")
   ;;(site-url :accessor site-url-of :initform "http://localhost:8080/")
   (site-url :accessor site-url-of :initform "http://lambda.s55.xrea.com/")
   (author :accessor author-of :initform "lambda_list@hotmail.com")
   ;;(template :accessor template-of :initform "default.html")
   (template :accessor template-of :initform "xrea.html")
   (rss-template :accessor rss-template-of :initform "default.rss")
   (logfile :accessor logfile-of :initform "ChangeLog.txt")
   ;; command
   (short    :accessor short-of :initform '((#\_ :link)
					    (#\/ :command)))
   ;; plugin
   (system-page :accessor system-page-of :initform (make-instance '<page>))
   (plugins :accessor plugins-of :initform nil)
   (included :accessor included-of :initform nil) ;; check recursive include
   ;; data
   (days   :accessor days-of  :initform nil)
   (pages  :accessor pages-of :initform nil)
   (files  :accessor files-of :initform nil)
   (system :accessor system-of :initform nil)))

(defmethod search-files ((self <weblog>) dir name type)
  (directory (make-pathname :directory (append (pathname-directory (src-of self)) (list dir))
			    :name name :type type)))

(defmethod list-files ((self <weblog>) dir pred fn)
  (let ((lst nil))
    (pathnames:walk-directory (merge-pathnames dir (src-of self))
			      (lambda (f)
				(if (funcall pred f) (push (funcall fn f) lst)))
			      :directories nil)
    (nreverse lst)))

(defmethod name-of ((self <file>))
  (pathname-name (path-of self)))

(defmethod filetype-of ((self <file>))
  (pathname-type (path-of self)))

(defmethod name-of ((self <day-page>))
  (let ((dir (pathname-directory (path-of self))))
    (format nil "~{~A~}~A" (subseq dir (- (length dir) 2)) (pathname-name (path-of self)))))

(defmethod update ((self <page>) contents &optional name summary)
  (let ((log (find-page (root-of self) :system (logfile-of (root-of self)))))
    (with-open-file (s (path-of self) :direction :output :if-exists :supersede)
      (write-sequence contents s))
    ;;(format t "name: ~S summary: ~S page: ~S~%" name summary log)
    (when (and name summary (not (string= name "")) (not (string= summary "")) log)
      (push (list (get-universal-time) (name-of self) name summary) (contents-of log))
      (with-open-file (s (path-of log) :direction :output :if-exists :supersede)
	(dolist (e (contents-of log))
	  (let ((str (with-output-to-string (s)
		       (format s "(~A \"~A\" \"~A\" \"~A\")"
			       (nth 0 e)
			       (nth 1 e)
			       (nth 2 e)
			       (nth 3 e)))))
	    (write-line str s)))))
    (setf (contents-of self) contents
	  (mtime-of self)    (file-write-date (path-of self)))))

(defmethod init ((self <file>))
  (let ((path (path-of self)))
    (setf (contents-of self) (read-file-contents path)
	  (mtime-of self)    (file-write-date path))))

(defmethod init ((self <page>))
  (let ((path (path-of self)))
    (setf (contents-of self) (read-file-contents path)
	  (mtime-of self)    (file-write-date path) )))

(defmethod init ((self <logfile>))
  (let ((path (path-of self)))
    (setf (contents-of self) nil)
    (with-open-file (s path :direction :input)
      (do ((line (read-line s nil :eof) (read-line s nil :eof)))
	  ((eq line :eof))
	(let ((lst (read-from-string line)))
	  (push lst (contents-of self))))) ;; XXX: need format check ?
    (setf (contents-of self) (nreverse (contents-of self)))
    (setf (mtime-of self)    (file-write-date path))))

(defparameter *plugin-document* nil)
(defparameter *plugin-function* nil)
(defmacro defplugin (doc fun)
  `(progn
    (setf *plugin-document* ,doc)
    (setf *plugin-function* ,fun)))

(defmethod init ((self <plugin>))
  (let ((path (path-of self)))
    (setf (contents-of self) (read-file-contents path)
	  (mtime-of self)    (file-write-date path))
    (let ((*compile-verbose* nil)
	  (*load-verbose* nil))
      (compile-file path)
      (load (make-pathname :directory (pathname-directory path)
			   :name (pathname-name path))))
    (setf (document-of self) *plugin-document*)
    (setf (function-of self) *plugin-function*)
    ;;(format t "~&Plugin: ~S loaded. (~A)~%" (name-of self) (function-of self))
    (format t "~&Plugin: ~A loaded. ~%" (name-of self))))

(defmethod init ((self <weblog>))
  (setf (days-of self) (list-files self "days"
				   (lambda (p) (string-equal (pathname-type p) "wiki"))
				   (lambda (p) (make-instance '<day-page> :path p :root self)))
	(pages-of self) (loop for file in (search-files self "pages" :wild "wiki")
			      collect (make-instance '<wiki-page> :path file :root self))
	(plugins-of self) (loop for file in (search-files self "plugins" :wild "lisp")
				collect (make-instance '<plugin> :path file :root self))
	(files-of self)  (list-files self "files"
				    #'identity
				    (lambda (p) (make-instance '<file> :path p :root self)))
	(system-of self) (loop for file in (search-files self "system" :wild :wild)
			      when (and (string= (pathname-name file) (pathname-name (logfile-of self)))
					(string= (pathname-type file) (pathname-type (logfile-of self))))
			      collect (make-instance '<logfile> :path file :root self)
			      else
			      collect (make-instance '<file> :path file :root self)))
  (setf (days-of self) (sort (days-of self) #'(lambda (e1 e2) (string> (name-of e1) (name-of e2)))))
  (loop for page in (pages-of self) do (init page))
  (loop for day in (days-of self) do (init day))
  (loop for data in (files-of self) do (init data))
  (loop for plugin in (plugins-of self) do (init plugin))
  (loop for system in (system-of self) do (init system))
  self)

(defmethod initialize-instance :after ((self <weblog>) &key)
  ;;(format *terminal-io* "~&weblog initialize ... ")
  (init self)
  (format *terminal-io* "blog initialize ... done.~%")
  self)

(defmethod find-page ((self <weblog>) (type (eql :page)) name)
  (or (loop for p in (pages-of self)
	    if (string-equal (name-of p) (pathname-name name))
	    do (return p))
      (loop for p in (days-of self)
	    if (string-equal (name-of p) (pathname-name name))
	    do (return p))))

(defmethod find-page ((self <weblog>) (type (eql :file)) name)
  (loop for p in (files-of self)
	if (let ((len (- (length (namestring (path-of p))) (length name))))
	     (and (> len 0) (string= (subseq (namestring (path-of p)) len) name)))
	do (return p)))

(defmethod find-page ((self <weblog>) (type (eql :plugin)) name)
  (loop for p in (plugins-of self)
	if (string-equal name (name-of p))
	do (return p)))

(defmethod find-page ((self <weblog>) (type (eql :system)) name)
  (loop for p in (system-of self)
	if (string-equal name (concatenate 'string (name-of p) "." (filetype-of p)))
	do (return p)))

(defmethod call-plugin ((self <weblog>) (req <request>) (page <page>) cmd stream)
  (let* ((name   (car cmd))
	 (arg    (cdr cmd))
	 (plugin (and name (find-page self :plugin name))))
    ;;(format t "[debug:call-plugin] ~A ~A ~A~%" name arg plugin)
    (if plugin
	(funcall (function-of plugin) self req page arg stream)
	(format stream "[Plugin Not Found: (~A, ~A)]" name arg))))

(defmethod dispatch-plugin ((self <page>) (req <request>)  token arg stream)
  (format stream "[Plugin Not Found: (~A, ~A)]" token arg))

(defmethod dispatch-plugin ((self <page>) (req <request>) (token (eql :link)) arg stream )
  ;; (format t "[debug:dispatch-plugin] link ~A~%" arg)
  (call-plugin (root-of self) req self (cons "link" arg) stream))

(defmethod dispatch-plugin ((self <page>) (req <request>)  (token (eql :command)) arg stream)
  ;;(format t ">>> arg: ~A~%" arg)
  (let* ((p (position #\Space arg))
	 (name (or (and p (subseq arg 1 p))
		   (subseq arg 1 (1- (length arg)))))
	 ;;(arg  (cdr (read-from-string arg))))
	 (lst (handler-case (read-from-string arg)
		(error (e) (declare (ignore e)) t))))
    ;; (format t "[debug:dispatch-plugin] ~A ~A~%"  name arg)
    (if (eq lst t)
	(format stream "[read error: ~A]" arg)
	(call-plugin (root-of self) req self (cons name (cdr lst)) stream))))

(defmethod %parse-line ((self <page>) (req <request>)  line)
  (with-output-to-string (stream)
    (labels ((dispatch (token arg)
	       (dispatch-plugin self req token arg stream))
	     (output (c)
	       (case c
		 (#\& (write-string "&amp;" stream))
		 (#\> (write-string "&gt;" stream))
		 (#\< (write-string "&lt;" stream))
	         (t (write-char c stream)))))
      (with-input-from-string (input line)
	(scan-stream (short-of (root-of self)) input #'output #'dispatch)))))

(defmethod parse-line ((self <page>) (req <request>)  line context)
  (labels ((%html-escape (line)
	     (with-output-to-string (out)
	       (loop for c across line do
		     (case c
		       (#\Tab (write-string "    " out))
		       (#\& (write-string "&amp;" out))
		       (#\> (write-string "&gt;" out))
		       (#\< (write-string "&lt;" out))
		       (t   (write-char c out))))))
	   (comment? (line context)
	     (and (not (eq context :varb))
		  (string=? ";;" line)))
	   (plugin? (line context)
	     (and (eq context :empty)
		  (string=? "/(" line)))
	   (item-list? (line context)
	     (and (not (eq context :varb))
		  (or (and (string=? "---- " line) :ul4)
		      (and (string=? "--- " line) :ul3)
		      (and (string=? "-- " line) :ul2)
		      (and (string=? "- " line) :ul1))))
	   (hr? (line context)
	     (and (not (eq context :varb))
		  (< (length line) 5)
		  (string=? "----" line)))
	   (headline? (line context)
	     (and (not (eq context :varb))
		  (or (and (string=? "**** " line) :h4)
		      (and (string=? "*** " line) :h3)
		      (and (string=? "** " line) :h2)
		      (and (string=? "* " line) :h1))))
	   (pre? (line context)
	     (and (not (eq context :varb))
		  (> (length line) 1) (eql (char line 0) #\Space)))
	   (begin-varbatim? (line context)
	     (and (not (eq context :varb))
		  (string=? "{{{" line)))
	   (end-varbatim? (line context)
	     (and (eq context :varb)
		  (string=? "}}}" line)))
	   (empty? (line context)
	     (and (not (eq context :varb))
		  (every #'(lambda (c) (find c '(#\Space #\Tab #\return #\linefeed))) line))))
    (acond
     ((comment? line context)
      (values (%html-escape line) :comment))
     ((plugin? line context)
      (values (%parse-line self req line) :empty))
     ((item-list? line context)
      (values (%parse-line self req
			   (ecase it
			     (:ul1 (subseq line 2))
			     (:ul2 (subseq line 3))
			     (:ul3 (subseq line 4))
			     (:ul4 (subseq line 5))))
	      :ul
	      (ecase it
		(:ul1 1) (:ul2 2) (:ul3 3) (:ul4 4))))
     ((hr? line context)
      (values line :hr))
     ((headline? line context)
      (let ((str (ecase it
		   (:h1 (subseq line 2))
		   (:h2 (subseq line 3))
		   (:h3 (subseq line 4))
		   (:h4 (subseq line 5))))
	    (level (ecase it (:h1 1) (:h2 2) (:h3 3) (:h4 4))))
	(when (and (option-of self) (assoc :toc (option-of self)))
	  (let* ((info (assoc :toc (option-of self))))
	    (when (> (cdr info) level) (loop while (> (cdr info) level) do (decf (cdr info)) (write-line "</ul></li>" (header-of self))))
	    (when (< (cdr info) level) (loop while (< (cdr info) level) do (incf (cdr info)) (write-line "<li class=\"none\"><ul>" (header-of self))))
	    (format (header-of self) "<li><a href=\"#~A\">~A</a></li>~%" (url-encode-string-only-display-char str) str)))
	(values str :h level)))
     ((pre? line context)
      (values line :pre))
     ((begin-varbatim? line context)
      (values line :varb))
     ((end-varbatim? line context)
      (values line :varb :end))
     ((empty? line context)
      (values line :empty))
     (t
      (if (eq context :varb)
	  (values (%html-escape line) :varb)
	  (values (%parse-line self req line) :p))))))

(defmethod wiki-to-html ((self <page>) (req <request>) input out)
  (labels ((output (x) (write-string x out)))
    (do ((context :body)
	 (previous :body)
	 (level 0)
	 (line (read-line input nil :eof) (read-line input nil :eof)))
	((eq line :eof)
	 (when (assoc :toc (option-of self))
	   (let ((level (cdr (assoc :toc (option-of self)))))
	     (loop while (> level 1) do (decf level) (write-line "</ul></li>" (header-of self)))
	     (write-line "</li></ul>" (header-of self))))
	 (cond ((eq context :ul)
		(loop while (>= level 1) do (decf level) (write-line "</li></ul>" out)))
	       ((not (find context '(:body :empty :h)))
		(format out "</~A>~%" context))))
      (multiple-value-bind (str ctxt option)
	  (parse-line self req line context)
	(setf previous context
	      context  ctxt)
	;;(format *terminal-io* "~S ~A ~A~%" str ctxt option)
	(when (and (eq previous :p) (not (eq context :p)))
	  (write-line "</p>" out))
	(cond ((and (not (eq previous :varb)) (eq context :varb))
	       (write-line "<pre>" out) (setq str nil))
	      ((and (eq context :varb) (eq option :end))
	       (write-line "</pre>" out)
	       (setq str nil context :body))
	      ((and (not (eq previous :ul)) (eq context :ul))
	       (write-string "<ul>" out)
	       (setf level 1))
	      ((and (eq previous :ul) (not (eq context :ul)))
	       (loop while (> level 1) do (decf level) (write-line "</li></ul></li>" out))
	       (write-line "</li></ul>" out)))
	(when str
	  (cond ((eq context :comment)
		 (setf context previous)
		 (when (and (search-str "toc" str) (not (assoc :toc (option-of self))))
		   (push (cons :toc 1) (option-of self))
		   (write-line "<ul>" (header-of self)))
		 (write-string "<!--" out) (output str) (write-line "-->" out))
		((eq context :hr)
		 (setf context previous)
		 (write-line "<hr />"))
		((eq context :ul)
		 (cond ((> level option) (loop while (> level option) do (decf level) (write-line "</li></ul></li>" out)))
		       ((< level option) (loop while (< level option) do (incf level) (write-line "<ul>" out)))
		       ((eq previous :ul) (write-line "</li>" out)))
		 (write-string "<li>" out) (output str))
		;;((eq context :h) (format out "<h~D>" option) (output str) (format out "</h~D>~%" option))
		((eq context :h)
		 (format out "<h~D><a id=\"~A\">~A</a></h~D>" option (url-encode-string-only-display-char str) str option))
		(t
		 (when (and (not (eq previous :p)) (eq context :p)) (write-string "<p>" out))
		 (output str) (write-line "" out))))))))

(defmethod make-renderd-page ((self <page>) (req <request>))
  (setf (option-of self) nil
	(header-of self) ""
	(body-of self)   (contents-of self)
	(footer-of self) ""))

(defmethod make-renderd-page ((self <wiki-page>) (req <request>))
  (let ((header (make-string-output-stream))
	(body   (make-string-output-stream))
	(footer (make-string-output-stream)))
    (setf (option-of self) nil
	  (header-of self) header
	  (body-of self)   body
	  (footer-of self) footer)
    (with-input-from-string (input (contents-of self))
      (wiki-to-html self req input body))
    (setf (header-of self) (get-output-stream-string (header-of self))
	  (body-of self)   (get-output-stream-string (body-of self))
	  (footer-of self) (get-output-stream-string (footer-of self)))))

(defmethod fill-template ((self <page>) (req <request>))
  (with-output-to-string (stream)
    (labels ((dispatch (token arg)
	       (dispatch-plugin self req token arg stream))
	     (output (c)
	       (write-char c stream)))
      (let ((template (find-page (root-of self) :system (template-of (root-of self)))))
	(if (null template) (error "Couldn't find template"))
	(make-renderd-page self req)
	(when template
	  (with-input-from-string (input (contents-of template))
	    (scan-stream (short-of (root-of self)) input #'output #'dispatch)))))))

(defmethod fill-template ((self <logfile>) (req <request>))
  (with-output-to-string (stream)
    (labels ((dispatch (token arg)
	       (dispatch-plugin self req token arg stream))
	     (output (c)
	       (write-char c stream)))
      (let ((template (find-page (root-of self) :system (rss-template-of (root-of self)))))
	(when template
	  (with-input-from-string (input (contents-of template))
	    (scan-stream (short-of (root-of self)) input #'output #'dispatch)))))))

(defmethod publish ((self <weblog>))
  (let ((lst (append (days-of self) (copy-list (pages-of self))))
	(req (make-instance '<request> :method :static))
	(log (find-page self :system (logfile-of self))))
    (dolist (page lst)
      (with-open-file (out (make-pathname :directory (pathname-directory (dest-of self))
					  :name      (name-of page)
					  :type      "html")
			   :direction :output
			   :if-exists :supersede)
        (setf (included-of self) nil) ;; clear include check
	;; (format *terminal-io* "~&>>> ~A (~A)~%" page (name-of page)) ; for debug
	(render-html self req out page :html)))
    (when log
      (with-open-file (out (make-pathname :directory (pathname-directory (dest-of self))
					  :name      "recent-changes"
					  ;;:type      "rdf")
					  :type      "rss")
			   :direction :output
			   :if-exists :supersede)
        (setf (included-of self) nil) ;; clear include check
	;; (format *terminal-io* "~&>>> ~A (~A)~%" log (name-of log)) ; for debug
	(render-html self req out log :file)))))

(defmethod render (self req s page cmd)
  (write-line "HTTP/1.0 200 OK" s)
  (write-line "Content-type: text/html" s)
  ;;(write-line "Content-type: application/xhtml+xml" s)
  (write-line "" s)
  (render-html self req s page cmd))

(defmethod render (self req s (page <file>) (cmd (eql :file)))
  (render-html self req s page cmd))

(defmethod render ((app <weblog>) (req <request>) s url (cmd (eql :redirect)))
  (write-line "HTTP/1.0 302 OK" s)
  (format s  "Location: ~A~%" url)
  (write-line "" s))

(defmethod render-html ((self <weblog>) (req <request>) s (page <file>) (cmd (eql :file)))
  (write-line "HTTP/1.0 200 OK" s)
  (cond ((string-equal (pathname-type (path-of page)) "css")
	 (write-line "Content-Type: text/css" s))
	((string-equal (pathname-type (path-of page)) "png")
	 (write-line "Content-Type: image/png" s))
	((string-equal (pathname-type (path-of page)) "html")
	 (write-line "Content-Type: text/html" s))
	(t
	 (write-line "Content-Type: text/plain" s)))
  (write-line "" s)
  (write-sequence (contents-of page) s))

(defmethod render-html ((self <weblog>) (req <request>) s (page <logfile>) (cmd (eql :file)))
  (unless (eq (method-of req) :static)
    (write-line "HTTP/1.0 200 OK" s)
    (write-line "Content-Type: text/xml" s)
    (write-line "" s))
  (write-line (fill-template page req) s))

(defmethod render-html ((self <weblog>) (req <request>) s (page <page>) (cmd (eql :html)))
  (write-line (fill-template page req) s))

#-cmu
(declaim (inline search-str))
#-cmu
(defun search-str (pattern str)
  (search pattern str))
#+cmu
(defun search-str (pattern str)
  (declare (simple-string pattern str)
	   (optimize (speed 3) (space 0) (safety 0)))
  (loop with len of-type fixnum  = (length pattern)
	with last of-type fixnum = (length str)
	for p1 of-type fixnum from 0 below (- last len) do
	(loop for p2 of-type fixnum from 0 to len
	      when (= p2 len) do (return-from search-str p1)
	      while (char-equal (schar pattern p2) (schar str (+ p1 p2))))))


(defmethod render-html ((self <weblog>) (req <request>) s arg (cmd (eql :search)))
  (let* ((name (subseq arg 0 (- (length arg) 5)))
	 (key  (subseq arg 7 (- (length arg) 5)))
	 (page (find-page self :page name)))
    (unless page
      (setf page (make-instance '<wiki-page>
				:root self
				:path (make-pathname :directory (pathname-directory (src-of self))
						     :name      name
						     :type      "html")))
      (push page (pages-of self)))
    (setf (mtime-of page) (get-universal-time)
	  (contents-of page)
	  (with-output-to-string (out)
	    (let* ((lst (split-string (url-decode-string key) #\Space))
		   (from  (cond ((or (null lst) (= (length lst) 1)) :default)
				((string= "pages" (first lst)) :pages)
				((string= "days" (first lst)) :days)
				(t :all)))
		   (src (sort (ecase from
				(:pages (copy-list (pages-of self)))
				(:days  (copy-list (days-of self)))
				((:all :default)
				 (append (days-of self) (copy-list (pages-of self)))))
			      #'string< :key #'name-of))
		   (words (ecase from
			    ((:all :pages :days) (cdr lst))
			    (:default lst))))
	      (format out "* Search ~S~%" words)
	      (loop with hits = 0
		    for p in src
		    when (and (not (string=? "search_" (name-of p)))
			      (every #'(lambda (w) (and (not (string= w ""))
							(search-str w (contents-of p)))) words))
		    do
		    (format out "- _(~A)~%" (name-of p))
		    (incf hits)
		    finally
		    (format out "search result: ~D hits~%" hits)))))
    (write-line (fill-template page req) s)))

(defmethod render-html ((self <weblog>) (req <request>) s (page <page>) (cmd (eql :edit)))
  (let ((system-page (system-page-of self)))
    (setf (root-of system-page) self
	  (path-of system-page) (path-of page)
	  (mtime-of system-page) (mtime-of page)
	  (contents-of system-page)
	  (with-output-to-string (out)
	    (format out "<h1>~A (Edit)</h1>~%" (name-of page))
	    (format out "<form method=\"post\" action=\"~A.html\"><p>~%" (name-of page))
	    (write-line "<input type=\"hidden\" name=\"edit\" value=\"true\" />" out)
	    (write-line "<textarea name=\"contents\" cols=\"70\" rows=\"25\">" out)
	    (write-sequence (contents-of page) out)
	    (write-line "</textarea><br />" out)
	    (write-line "<b>Summary of change:</b><input type=\"text\" size=\"60\" name=\"summary\" value=\"\" /><br />" out)
	    (write-line "<b>Your name:</b><input type=\"text\" size=\"30\" name=\"name\" value=\"lambda\" /><br />" out)
	    (write-line "<input type=\"submit\" value=\"Commit\" /></p></form>" out)))
    (write-line (fill-template system-page req) s)))

(defmethod render-html ((self <weblog>) (req <request>) s name (cmd (eql :notfound)))
  (let ((system-page (system-page-of self)))
    (setf (root-of system-page) self
	  ;;(path-of system-page) (path-of page)
	  ;;(mtime-of system-page) (mtime-of page)
	  (contents-of system-page)
	  (with-output-to-string (out)
	    (format out "<h1>~A (Not Found)</h1>" name)
	    (format out "<p>404 Not Found.</p>")
	    (format out "<form method=\"post\" action=\"~A.html\"><p>" name)
	    (write-line "<input type=\"hidden\" name=\"create\" value=\"page\" />" out)
	    (write-line "<input type=\"submit\" value=\"Create Page\" /></p></form>" out)))
    (write-line (fill-template system-page req) s)))

(defmethod page-view ((app <weblog>) (req <request>) s)
  (let* ((local? (localhost? req))
	 (edit?     (param-check req :param "edit" "true"))
	 (contents? (param-check req :param "contents" (lambda (e) (and (stringp e) e))))
	 (create?   (param-check req :param "create" (lambda (e) (and (stringp e) e))))
	 (search?   (param-check req :param "search" (lambda (e) (and (stringp e) e))))
	 (lst (split-string (url-of req) #\/ :start 1))
	 (cmd (car lst))
	 (arg (cdr lst)))
    (setf (included-of app) nil) ;; clear include check
    ;;(format t "[debug] cmd: ~S arg: ~S~%" cmd arg)
    (cond ((and local? (string-equal cmd "initialize"))
	   (write-string "server initialize" *terminal-io*)
	   (init app)
	   (write-line "... done." *terminal-io*)
	   (render app req s "index.html" :redirect))
	  ((and local? (string-equal cmd "publish"))
	   (render app req s "index.html" :redirect)
	   (publish app))
	  ((or (string-equal cmd "recent-changes.rdf")
	       (string-equal cmd "recent-changes.rss"))
	   (let ((page (find-page app :system (logfile-of app))))
	     (if page
		 (render app req s page :file)
		 (render app req s "index.html" :redirect))))
	  ((string-equal cmd "files")
	   (let ((name #-WIN32(format nil "files~{/~A~}" arg)
	               #+WIN32(format nil "files~{\\~A~}" arg)))
	     (if (find-page app :file name)
		 (render app req s (find-page app :file name) :file)
		 (render app req s "../index.html" :redirect))))
	  ;;((and cmd arg)
	  ;; (render app req s "../index.html" :redirect))
	  ((string= cmd "")
	   (render app req s "index.html" :redirect))
	  ;; support search
	  ((or search? (string=? "search_" cmd))
	   (if search?
	       (let* ((from (param-check req :param "from" (lambda (e) (and (stringp e) e))))
		      (type (cond ((string=? from "pages") "pages")
				  ((string=? from "days") "days")
				  (t "all"))))
		 ;;(format t "debug ~A, ~A~%" search? from)
		 (render app req s (format nil "search_~A+~A.html" type search?) :redirect))
	       (render app req s cmd :search)))
	  (t
	   (let ((page (find-page app :page cmd)))
	     (cond ((and (null page)
			 create?
			 (string-equal create? "dayfile")
			 (and (>= (length cmd) 8)
			      (every #'digit-char-p (subseq cmd 0 8))))
		    (let* ((year (subseq cmd 0 4))
			   (month (subseq cmd 4 6))
			   (day   (subseq cmd 6 8))
			   (page  (make-instance '<day-page>
						 :root app
						 :mtime (get-universal-time)
						 :path (make-pathname :directory
								      (append (pathname-directory (src-of app))
									      (list "days" year month))
								      :name day :type "wiki"))))
		      (create-directories (path-of page))
		      (push page (days-of app))
		      (render app req s page :edit)))
		   ((and (null page)
			 create?
			 (string-equal create? "page"))
		    (let* ((pos (position #\. cmd))
			   (name (subseq cmd 0 pos))
			   (page (make-instance '<wiki-page>
						:root app
						:mtime (get-universal-time)
						:path (make-pathname :directory
								     (append (pathname-directory (src-of app))
									     '("pages"))
								     :name name :type "wiki"))))
		      (push page (pages-of app))
		      (render app req s page :edit)))
		   ((null page)
		    (render app req s cmd :notfound))
		   ((and edit? contents?)
		    (let ((name?     (param-check req :param "name" (lambda (e) (and (stringp e) e))))
			  (summary?  (param-check req :param "summary" (lambda (e) (and (stringp e) e)))))
		      (update page contents? name? summary?)
		      (render app req s (name-of page) :redirect)))
		   ((and edit?)
		    (render app req s page :edit))
		   (page
		    (render app req s page :html))))))))

(defwebapp <weblog> :name "Common Lisp Weblog" :base-url "/")
